//
//  main.cpp
//  rttfp
//
//  Created by Stanko Juzbasic on 1/19/16.
//
//  this is the Text File Player program soruce code , written in RTaudio API

#include "RtAudio.h"
#include <iostream>
#include <cstdlib>
#include <cmath>
#include <cstring>
#include <unistd.h>
#include <sys/select.h>


//____declarations_______________________________________________________________________
// globals
unsigned int sampleRate = 44100;

RtAudio dac;

typedef struct {
    double  j;
    unsigned int   N;
    unsigned int   totalFrames;
    float*         fBuffer;
    size_t         fbuffersize;
    bool           STOP;
}uData;


//Prototypes______________________________________________________________________________

double kramp(double);

void   usage (void);

size_t getSize(FILE* , char* );

void normalize(float* , size_t , float );

int fbrp(void*, void*, unsigned int, double, RtAudioStreamStatus, void* ); //rendering
//________________________________________________________________________________________
#pragma mark - auxilliary function(s) -
//________________________________________________________________________________________
void usage (void)
{
    std::cout << "\nuseage: rttfp fs file <device> <channelOffset>\n";
    std::cout << "    fs = the sample rate, \n";
    std::cout << "    file = the raw file to play,\n";
    std::cout << "    device = optional device to use (default = 0),\n";
    std::cout << "    and channelOffset = an optional channel offset on the device (default = 0).\n\n";
    return;
}
//________________________________________________________________________________________
size_t getSize(FILE* file, char* name)
{
    size_t s;
    if(file == NULL)return 0;
    fseek(file, 0L, SEEK_END);
    s = ftell(file);
    s = s/sizeof(float);
    std::cout<<"File "<<name<<": "<<s<<"\t values\n"<<std::endl;
    rewind(file);
    return s;
}
//________________________________________________________________________________________
void normalize(float* buf, size_t siz, float norm_max)
{
    float m;
    if(norm_max > .99f)norm_max = .99f;
    for (int i = 0; i < siz; i++)
        if (fabsf(buf[i]) > m ) {
            m = fabsf(buf[i]);
        }
    m = norm_max/m;
    for (int j = 0 ; j < siz; j++)
        buf[j] = m*buf[j];
    
    return;
}
//___________________________________________________________________________________
// generate 1000 ramp samples (it's a hack!)
double kramp ( double i)
{
    if(i >= 1000.)return 1.;
    if(i <= 0.)   return 0.;
    return .001*(i);
}

//____________________________________________________________________________________
#pragma mark - callback -
//____________________________________________________________________________________
// Two-channel float buffer player with start/stop
int fbrp(void *outputBuffer,
         void *inputBuffer,
         unsigned int nBufferFrames,
         double streamTime,
         RtAudioStreamStatus status,
         void *userData )
{

    double *buffer = (double *) outputBuffer;

    uData *data = (uData*)userData;
    
    unsigned int frame;
    double  d;
    long   l = (long)data->j;
    
    if ( status )
        std::cout << "Stream underflow detected!" << std::endl;
    
    // Writing interleaved stereo audio data from float buffer
    for (frame = 0; frame < nBufferFrames; ++frame)
	{
        d = 0.;

        //copy data to buffer
        if(data->totalFrames >= data->fbuffersize){
            d =  0.;
        }else if(data->STOP){
            if(data->N < 1)data->N = 1;
            d =  data->fBuffer[l]*kramp((double)data->N--);
           
        }else{
            d =  data->fBuffer[l];
        }
        
        buffer[2*frame+1] = buffer[2*frame] = (double)d;

		data->j += 1.0;
        data->totalFrames +=1;
        l++;
        if(0)std::cout<<data->N<<std::endl;
	}

    if(data->totalFrames >= data->fbuffersize){
        std::cout<<"game over!"<<std::endl;
        //signal the stream to stop by passing a non-zero return value
        return 1;
    }
    return 0;
}
//____________________________________________________________________________________
#pragma mark - main -
//____________________________________________________________________________________
int main(int argc, const char * argv[])
{
    FILE *afPtr;
    char afname[512];
    char text  [16];
    size_t size;
    
    unsigned int  fs, bufferFrames, device = 0, offset = 0;
    
    int fe;

    uData data;
    
    data.N = 2000; data.STOP = false;
    
    if((argc < 3)||(argc > 5)){
        usage();
        exit (0);
    }
    
    std::cout << "\n_____________________________________________\n";
    std::cout << "* * * * Text File Player  for RTAudio * * * *\n";    
    std::cout <<"\n\n";
    
    if(0)    std::cout << argc <<std::endl;
    
    //parsing command-line arguments:
    fs = (unsigned int) atoi( argv[1] );
    if (argv[2] == NULL)
        goto no_name;
    strcpy(afname, argv[2]);
    std::cout<<"\nName: "<<argv[2]<<"\tFile: "<<afname<<std::endl;
    if ( argc > 3 )
        device = (unsigned int) atoi( argv[3] );
    if ( argc > 4 )
        offset = (unsigned int) atoi( argv[4] );
    
    //getting the file, its size and content, allocating buffer:
    afPtr = fopen(afname,"rb");
    
    if (afPtr == NULL){
        std::cout<<"File \""<<argv[2]<<"\"\tnot found"<<std::endl;
    no_name:
        usage();
        exit(0);
    }
    
    //size check
    if((size = getSize(afPtr, afname)) <= 0){
        std::cout<<"File "<<afname<<" is empty"<<std::endl;
        fclose(afPtr);
        exit(0);
    };
    
    //actual size read
    size = 1;
    while( !feof(afPtr)) {
        fscanf(afPtr, "%s",text);
        size++;
    }    
    rewind(afPtr);
    
    
    if((data.fBuffer  = (float*)calloc(size + 1000, sizeof(float))) == NULL){
        std::cout<<"Not enough memory!"<<std::endl;
        fclose(afPtr);
        exit(0);
    }
    
    data.fbuffersize = size;
        
    //fread(data.fBuffer,sizeof(float),size, afPtr);
    
    int i = 0;
    while( !feof(afPtr)) {
        fscanf(afPtr, "%s",text);
        float val = (float)atof(text);
        data.fBuffer[i++] = val;
    }
    
    normalize(data.fBuffer, size, 0.9f);
    
    fclose(afPtr);
        
   

    //Open RTaudio stream

    if ( dac.getDeviceCount() < 1 ) {
        std::cout << "\nNo audio devices found!"<<std::endl;
        exit (0);
    }
    
    
    if(0){
    std::cout << "option "<<fe<<std::endl;
            exit(-1);
    }

    
    RtAudio::StreamParameters parameters;
    parameters.deviceId = dac.getDefaultOutputDevice();
    parameters.nChannels = 2;
    parameters.firstChannel = 0;
    data.totalFrames = 0;
    bufferFrames = 512; // 512 sample frames
    
    data.j = 0.; 
    try {
        dac.openStream( &parameters, NULL, RTAUDIO_FLOAT64,
                       sampleRate, &bufferFrames, &fbrp, (void *)&data );
        dac.startStream();
    }
    catch ( RtAudioError& e ) {
        e.printMessage();
        exit( 0 );
    }
    

    //stop when over, but also when interrupted by pressing <enter> !
    struct timeval tv;
    tv.tv_sec =(time_t)(data.fbuffersize/(double)sampleRate);
    tv.tv_usec=(int)(1000000*((data.fbuffersize/(double)sampleRate) - tv.tv_sec));
    std::cout<<"Duration: "<<tv.tv_sec<<"."<<tv.tv_usec<<"s"<<std::endl;
    std::cout << "\nPlaying ... press <enter> to stop."<<std::endl;
    
    fd_set read_fd;
    FD_ZERO(&read_fd);
    FD_SET(STDIN_FILENO, &read_fd);
    
    if (select(STDIN_FILENO+1, &read_fd, NULL, NULL, &tv) < 0) std::cerr<<"select"<<std::endl;
    
    bool b = (FD_ISSET(0,&read_fd)) ;
    if(b==1){
        std::cout<<"Playback stopped..."<<std::endl;
        data.STOP = true;
    }
    else
        std::cout<<"Playback finished..."<<std::endl;

    std::cout<<"Ending and quitting...\n"<<std::endl;
    
    //wait for the stopped playback to die out (tail)
    if (data.STOP)usleep(50000);
    
    if ( dac.isStreamRunning() == true )
        try {
        // Stop the stream    
        dac.stopStream();
        }
        catch (RtAudioError& e) {
            e.printMessage();
        }
    
cleanup:    
    if ( dac.isStreamOpen() ) dac.closeStream();
    return 0;
}

